#!/usr/bin/perl

use strict;
use warnings;

print "Hello!\n\n";
print "Thank you for interest in FastNetMon Community\n";
print "We've moved to new installer tool documented here: https://fastnetmon.com/install/\n";

print "\n";

print "To start installation procedure please do following: \n\n";
print "wget https://install.fastnetmon.com/installer -Oinstaller\n";
print "sudo chmod +x installer\n";
print "sudo ./installer -install_community_edition\n";

print "\n";

